@@ -10,7 +10,7 @@ import tkinter.messagebox as messagebox
 import tkinter.ttk as ttk
 from pathlib import Path
 
-from natsort import natsorted
+import natsort as ns
 from PIL import Image, ImageTk
 
 logger = logging.getLogger(__name__)
@@ -69,6 +69,8 @@ class ArchiveBase:
 
         self.cache = {}
 
+        self.is_directory = False
+
     def __del__(self):
         self.close()
 
@@ -81,8 +83,8 @@ class ArchiveBase:
 
     def close(self):
         self.stop = True
-        if self.data is not None:
-            self.data.close()
+        # if self.data is not None:
+        #     self.data.close()
         self.file_path = None
         self.file_list = []
 
@@ -123,6 +125,7 @@ class ArchiveBase:
 
             if len(yet) == 0:
                 logger.debug("cache is full.")
+                logger.debug(f"file_path = {self.file_path}")
                 logger.debug(f"cached page is {self.cache.keys()}")
                 time.sleep(1)
                 continue
@@ -151,12 +154,12 @@ class ArchiveBase:
 
     def __getitem__(self, i):
         i = self.in_range(i)
-        self.i = i
 
         if self.cache.get(i) is not None:
             logger.debug("cache hit")
             return self.cache[i]
 
+        self.i = i
         logger.debug(f"cache failed:{i}")
         file_name, data = self.getitem(i)
         self.cache[i] = (file_name, data)
@@ -187,31 +190,38 @@ class ArchiveBase:
     def current(self):
         return self[self.i]
 
-    def trash(self):
-        logger.debug("called")
-        file_path = self.file_path
-        self.close()
-        if file_path is not None:
-            if "send2trash" not in globals():
-                global send2trash
-                from send2trash import send2trash
-            send2trash(str(file_path))
+    def sort_file_list(self):
+        self.file_list = ns.natsorted(
+            self.file_list, key=lambda x: str(x), alg=ns.PATH | ns.IGNORECASE
+        )
 
 
 class DirectoryArchive(ArchiveBase):
     def __init__(self, file_path, data=None):
         super().__init__()
+        self.is_directory = True
         self.open(file_path, data)
-        self.start_preload()
+        self.cache = {}
+        # self.start_preload()
+
+    def search(self, file_path):
+        self.i = self.file_list.index(Path(file_path))
+        self.cache = {}
+        return self.i
+
+    def remove(self, file_path):
+        i = self.search(file_path)
+        logger.debug(f"remove {i}:{file_path}")
+        self.cache = {}
+        del self.file_list[i]
 
     def open(self, file_path, data=None):
         # you cannot path data, ignored
         self.file_path = Path(file_path)
-        self.file_list = natsorted(
-            list(Path(self.file_path.parent).glob("*")), key=lambda x: str(x)
-        )
+        self.file_list = list(Path(self.file_path.parent).glob("*"))
+        self.sort_file_list()
         self.filtering_file_list()
-        logger.debug(self.file_list)
+        # logger.debug(self.file_list)
         self.i = self.file_list.index(Path(file_path))
         logger.debug(f"self.i = {self.i}")
 
@@ -223,9 +233,14 @@ class DirectoryArchive(ArchiveBase):
         return start, end, self.file_list[start:end], [None] * (end - start)
 
     def getitem(self, i):
+        # disable cache because trash not works well.
+        # Other way is is_directory and set file_path as same
+        self.cache = {}
         logger.debug(f"i = {i}")
         if 0 <= i < len(self):
-            return self.file_list[i], None
+            self.i = i
+            self.file_path = self.file_list[i]
+            return self.file_path, None
         else:
             return "", None
 
@@ -246,13 +261,17 @@ class ZipArchive(ArchiveBase):
         self.file_list = []
 
         logger.debug("to byte")
-        fp = self.file_path if data is None else io.BytesIO(self.data)
+        # fp = self.file_path if data is None else io.BytesIO(self.data)
+        if self.data is not None:
+            self.data.seek(0)
+        fp = self.file_path if data is None else self.data
 
         logger.debug("zip open")
         with zipfile.ZipFile(fp) as f:
-            self.file_list = natsorted(f.namelist())
-
+            self.file_list = f.namelist()
+            # ns.natsorted(f.namelist())
         logger.debug("to list")
+        self.sort_file_list()
         self.filtering_file_list()
         logger.debug(self.file_list)
         logger.debug("return")
@@ -263,7 +282,8 @@ class ZipArchive(ArchiveBase):
         file_byte = None
 
         logger.debug("to byte")
-        fp = self.file_path if self.data is None else io.BytesIO(self.data)
+        # fp = self.file_path if self.data is None else io.BytesIO(self.data)
+        fp = self.file_path if self.data is None else self.data
 
         logger.debug("open zip")
         if 0 <= i < len(self):
@@ -294,13 +314,16 @@ class RarArchive(ArchiveBase):
         self.file_list = []
 
         logger.debug("to byte")
-        fp = self.file_path if data is None else io.BytesIO(self.data)
+        if self.data is not None:
+            self.data.seek(0)
+        fp = self.file_path if data is None else self.data
 
         logger.debug("open rar")
         with rarfile.RarFile(fp) as f:
-            self.file_list = natsorted(f.namelist())
+            self.file_list = f.namelist()
 
         logger.debug("open rar")
+        self.sort_file_list()
         self.filtering_file_list()
         logger.debug(self.file_list)
 
@@ -309,7 +332,7 @@ class RarArchive(ArchiveBase):
         file_name = ""
         file_byte = None
         logger.debug("to byte")
-        fp = self.file_path if self.data is None else io.BytesIO(self.data)
+        fp = self.file_path if self.data is None else self.data
         logger.debug("read file")
         if 0 <= i < len(self):
             with rarfile.RarFile(fp) as f:
@@ -339,18 +362,23 @@ class SevenZipArchive(ArchiveBase):
         self.data = data
         self.file_list = []
         logger.debug("to bytes")
-        fp = self.file_path if self.data is None else io.BytesIO(self.data)
+        if self.data is not None:
+            self.data.seek(0)
+        fp = self.file_path if self.data is None else self.data
         logger.debug("open 7z")
         with py7zr.SevenZipFile(fp, mode="r") as f:
-            self.file_list = natsorted(f.getnames())
+            self.file_list = f.getnames()
 
+        self.sort_file_list()
         self.filtering_file_list()
         logger.debug(self.file_list)
         logger.debug("return")
 
     def getitems(self, start, end):
 
-        fp = self.file_path if self.data is None else io.BytesIO(self.data)
+        if self.data is not None:
+            self.data.seek(0)
+        fp = self.file_path if self.data is None else self.data
 
         file_names = []
         file_bytes = []
@@ -411,20 +439,6 @@ class PdfArchive(ArchiveBase):
         self.open(file_path, data)
         self.start_preload()
 
-    # this may eat too much memory
-    def open_all(self, file_path, data=None):
-        self.file_path = file_path
-
-        self.data = data
-
-        if data is None:
-            self.images = pdf2image.convert_from_path(file_path)
-        else:
-            self.images = pdf2image.convert_from_bytes(data.read())
-
-        page_num = len(self.images)
-        self.file_list = [Path(str(i + 1) + ".png") for i in range(page_num)]
-
     def open(self, file_path, data=None):
 
         self.file_path = file_path
@@ -504,6 +518,83 @@ class PdfArchive(ArchiveBase):
         return file_name, image
 
 
+class ArchiveTree:
+    def __init__(self):
+        self.root = []
+        pass
+
+    def reset(self):
+        self.root = []
+
+    def append(self, archive):
+        if archive is None:
+            logger.debug("archive is None.")
+            return
+        if archive.is_directory:
+            archive.stop = True
+            # self.root = [archive]
+            return
+
+        if len(self.root) > 1 and self.root[-1].file_path == archive.file_path:
+            logger.debug("same archive. skipping")
+            return
+        # do not delete file_path and file_list
+        # and stop preload because do not use so much.
+        archive.stop = True
+        self.root.append(archive)
+
+    def top(self):
+        if len(self.root) == 0:
+            return None
+        return self.root[0]
+
+    def next_archive(self):
+        num = len(self.root) - 1
+        if num < 0:
+            logger.debug("num < 0")
+            return "", None, None
+
+        for i in range(num, -1, -1):
+            archive = self.root[i]
+            file_path = archive.file_path
+            next_file_path, data = archive.next()
+            logger.debug(
+                f"i,file_path,next_file_path = {i},{file_path},{next_file_path}"
+            )
+            if file_path == next_file_path:
+                logger.debug("go to parent")
+                del self.root[i]
+                continue
+
+            # archive.start_preload()
+            logger.debug(f"next_file_path = {next_file_path}")
+            return next_file_path, data, archive
+
+        logger.debug("not found")
+        return "", None, None
+
+    def prev_archive(self):
+        num = len(self.root) - 1
+        if num < 0:
+            return "", None, None
+
+        for i in range(num, -1, -1):
+            archive = self.root[i]
+            file_path = archive.file_path
+            next_file_path, data = archive.prev()
+            logger.debug(
+                f"i,file_path,next_file_path = {i},{file_path},{next_file_path}"
+            )
+            if file_path == next_file_path:
+                logger.debug("go to parent")
+                del self.root[i]
+                continue
+            return next_file_path, data, archive
+
+        logger.debug("not found")
+        return "", None, None
+
+
 class ImageFrame(tk.Canvas):
     algorithm = {
         "Nearest": Image.NEAREST,
@@ -872,7 +963,7 @@ Reload      = r
 
 # a = /home/GuiltyCat/images/fantastic
 # b = /home/GuiltyCat/images/bravo
-# b = /home/GuiltyCat/images/wonderful
+# c = /home/GuiltyCat/images/wonderful
 #
 """
 
@@ -938,6 +1029,10 @@ class SaltViewer(tk.Tk):
 
         self.archive = None
 
+        # this directory is the parent of first file_path that passed
+        # toself.open()
+        self.root_dir = None
+
         logger.debug("set title")
         self.title("SaltViewer")
 
@@ -982,38 +1077,51 @@ class SaltViewer(tk.Tk):
 
         self.num = 0
 
+        self.tree = ArchiveTree()
+
         self.load_config(args)
 
     def move_file(self, event):
 
+        fullscreen = self.attributes("-fullscreen")
+        self.attributes("-fullscreen", False)
+
         logger.debug("called")
 
         file_path = self.archive.file_path
+        top = self.tree.top()
+        if top is not None:
+            file_path = Path(top.file_path)
+
         move_to_list = self.config.move_to_list
         logger.debug(f"{file_path}, {move_to_list}")
 
-        archive = DirectoryArchive(file_path)
-        if len(archive) == 0:
-            self.archive.close()
-            archive.close()
-            self.quit(None)
+        if self.root_dir is None:
+            logger.debug("Directory Archive")
+            self.root_dir = DirectoryArchive(file_path)
+            self.root_dir.stop = True
 
-            if not MoveFile().move_file(move_to_list, file_path):
-                logger.debug("move failed")
-                return
-            return
-
-        next_file_path = archive.next()[0]
-        if next_file_path == file_path:
-            next_file_path = archive.prev()[0]
-
-        archive.close()
         if not MoveFile().move_file(move_to_list, file_path):
             logger.debug("move failed")
+            self.attributes("-fullscreen", fullscreen)
             return
+
+        if len(self.root_dir) == 1:
+            self.archive.close()
+            self.quit(None)
+            return
+
         self.archive.close()
+        self.archive = None
+        self.tree.reset()
+
+        self.root_dir.remove(file_path)
+        self.root_dir.cache = {}
+        next_file_path = self.root_dir.next()[0]
+
+        self.attributes("-fullscreen", fullscreen)
+        logger.debug(f"open {next_file_path}")
         self.open(next_file_path)
-        pass
 
     def reload(self, event):
         self.archive.cache = {}
@@ -1040,7 +1148,6 @@ class SaltViewer(tk.Tk):
 
     def load_config(self, args):
         logger.debug("called")
-
         logger.debug("overwrite settings")
         for k, v in args.items():
             if v is None:
@@ -1115,24 +1222,59 @@ class SaltViewer(tk.Tk):
         self.current_page()
 
     def next_archive(self, event):
-        file_path = self.archive.file_path
+
+        if self.root_dir is None:
+            logger.debug("Directory Archive")
+            self.root_dir = DirectoryArchive(self.archive.file_path)
+            self.root_dir.stop = True
+
+        if self.archive.is_directory:
+            self.next_page(event)
+            return
+
+        next_file_path, data, archive = self.tree.next_archive()
+        if next_file_path in ["", self.archive.file_path]:
+            top = self.tree.top()
+            if top is not None:
+                self.root_dir.search(top.file_path)
+            self.tree.reset()
+            self.archive.close()
+            self.archive = None
+            file_path = self.root_dir.next()[0]
+            self.open(file_path)
+            return
+
         self.archive.close()
-        archive = DirectoryArchive(file_path)
-
-        file_path = archive.next()[0]
-        archive.close()
-
-        self.open(file_path)
+        self.archive = None
+        logger.debug(f"next_file_path = {next_file_path}")
+        self.open(next_file_path, data)
 
     def prev_archive(self, event):
-        file_path = self.archive.file_path
+        if self.root_dir is None:
+            logger.debug("Directory Archive")
+            self.root_dir = DirectoryArchive(self.archive.file_path)
+            self.root_dir.stop = True
+
+        if self.archive.is_directory:
+            self.prev_page(event)
+            return
+
+        next_file_path, data, archive = self.tree.prev_archive()
+        if next_file_path in ["", self.archive.file_path]:
+            top = self.tree.top()
+            if top is not None:
+                self.root_dir.search(top.file_path)
+            self.tree.reset()
+            self.archive.close()
+            self.archive = None
+            file_path = self.root_dir.prev()[0]
+            self.open(file_path)
+            return
+
         self.archive.close()
-        archive = DirectoryArchive(file_path)
-
-        file_path = archive.prev()[0]
-        archive.close()
-
-        self.open(file_path)
+        self.archive = None
+        logger.debug(f"next_file_path = {next_file_path}")
+        self.open(next_file_path, data)
 
     def construct_gui(self):
         logger.debug("called")
@@ -1146,29 +1288,58 @@ class SaltViewer(tk.Tk):
 
         dummy_img = Image.new("RGB", (10, 10), color="black")
         self.image.mode = "Raw"
+        logger.debug("----------------------------------")
+        logger.debug("dummy image")
+        logger.debug("----------------------------------")
         self.image.display(dummy_img)
         logger.debug("return")
 
     def trash(self, event):
-        if messagebox.askokcancel("Trash file?", "Trash file?"):
-            file_path = self.archive.file_path
-            print(f"Trash {file_path}")
-            archive = DirectoryArchive(file_path)
-            if len(archive) == 1:
-                print("Archive is empty")
-                self.archive.trash()
-                archive.close()
+
+        fullscreen = self.attributes("-fullscreen")
+        self.attributes("-fullscreen", False)
+
+        file_path = self.archive.file_path
+        logger.debug(f"file_path = {file_path}")
+        top = self.tree.top()
+        if top is not None:
+            file_path = Path(top.file_path)
+        logger.debug(f"file_path = {file_path}")
+        if messagebox.askokcancel("Trash file?", f"Trash file?\n{file_path}"):
+            global send2trash
+            from send2trash import send2trash
+
+            if self.root_dir is None:
+                self.root_dir = DirectoryArchive(file_path)
+                self.root_dir.stop = True
+
+            # logger.debug(f"root_dir file_list = {self.root_dir.file_list}")
+            self.root_dir.remove(file_path)
+            # logger.debug(f"root_dir file_list = {self.root_dir.file_list}")
+            send2trash(str(file_path))
+
+            if len(self.root_dir) == 0:
+                logger.debug("directory is empty")
                 self.quit(None)
                 return
-            next_file_path = archive.next()[0]
-            if next_file_path == file_path:
-                next_file_path = archive.prev()[0]
-            archive.close()
-            self.archive.trash()
+
             self.archive.close()
-            self.open(next_file_path)
+            self.archive = None
+
+            self.root_dir.cache = {}
+            next_file_path, data = self.root_dir.current()
+
+            if next_file_path == "":
+                next_file_path, data = self.root_dir.prev()
+
+            self.tree.reset()
+            logger.debug(f"next_file_path = {next_file_path}")
+            self.attributes("-fullscreen", fullscreen)
+
+            self.open(next_file_path, data)
         else:
-            print("Cancelled")
+            self.attributes("-fullscreen", fullscreen)
+            logger.debug("Cancelled")
 
     def toggle_page_mode(self, event):
         self.double_page = not self.double_page
@@ -1190,6 +1361,9 @@ class SaltViewer(tk.Tk):
             image2 = self._open_next()
             # back to current
             self.archive.prev()
+        logger.debug("----------------------------------")
+        logger.debug("current")
+        logger.debug("----------------------------------")
         self.image.display(image, image2, self.right2left)
 
     def _open_next(self, c=1):
@@ -1198,6 +1372,7 @@ class SaltViewer(tk.Tk):
         if file_path == "":
             logger.debug("file_path is empty")
             return None
+        logger.debug(f"file_path={file_path}")
         return self.open_file(file_path, data)
 
     def next_page(self, event):
@@ -1214,6 +1389,9 @@ class SaltViewer(tk.Tk):
             image2 = self._open_next()
             # in order to set index as first page
             self.archive.prev()
+        logger.debug("-------------------------------------")
+        logger.debug("next_page")
+        logger.debug("-------------------------------------")
         self.image.display(image, image2, self.right2left)
 
     def _open_prev(self, c=1):
@@ -1221,6 +1399,8 @@ class SaltViewer(tk.Tk):
         if file_path == "":
             logger.debug("file_path is empty")
             return None
+        logger.debug(f"file_path={file_path}")
+        logger.debug(f"self.archive.file_path={self.archive.file_path}")
         return self.open_file(file_path, data)
 
     def prev_page(self, event):
@@ -1230,38 +1410,59 @@ class SaltViewer(tk.Tk):
         image2 = None
         if self.double_page:
             image2 = self._open_prev()
+        logger.debug("-------------------------------------")
+        logger.debug("prev_page")
+        logger.debug("-------------------------------------")
         self.image.display(image, image2, not self.right2left)
 
     def quit(self, event):
-        self.archive.close()
+        if self.archive is not None:
+            self.archive.close()
         self.destroy()
 
     def open(self, file_path, data=None):
         if self.archive is not None:
-            self.archive.close()
+            self.archive.stop = True
         self.archive = self.open_archive(file_path, data)
         file_path, data = self.archive.current()
         image = self.open_file(file_path, data)
+        logger.debug("-------------------------------------")
+        logger.debug("open")
+        logger.debug(image)
+        logger.debug("-------------------------------------")
         self.image.display(image)
+        return image
 
     def open_archive(self, file_path, data=None):
         logger.debug("called")
         suffix = Path(file_path).suffix.lower()
         if suffix == ".zip":
             logger.debug("zip")
-            return ZipArchive(file_path, data)
+            archive = ZipArchive(file_path, data)
         elif suffix == ".rar":
             logger.debug("rar")
-            return RarArchive(file_path, data)
+            archive = RarArchive(file_path, data)
         elif suffix == ".7z":
             logger.debug("7z")
-            return SevenZipArchive(file_path, data)
+            archive = SevenZipArchive(file_path, data)
         elif suffix == ".pdf":
             logger.debug("pdf")
-            return PdfArchive(file_path, data)
+            archive = PdfArchive(file_path, data)
         else:
             logger.debug("directory")
-            return DirectoryArchive(file_path, data)
+            archive = DirectoryArchive(file_path, data)
+
+        # in the case of nested archive
+        if self.archive is None:
+            return archive
+
+        # logger.debug(f"self.archive file_list = {self.archive.file_list}")
+        # logger.debug(f"archive file_list = {archive.file_list}")
+        # if not self.archive.is_directory and not archive.is_directory:
+        #     logger.debug("self.tree.append")
+        self.tree.append(self.archive)
+
+        return archive
 
     def open_file(self, file_path, data=None):
         file_path = Path(file_path)
@@ -1269,7 +1470,9 @@ class SaltViewer(tk.Tk):
 
         logger.debug("set title")
         title = f"{file_path}:"
-        if self.archive.file_path.stem != str(file_path.parent):
+        if self.archive.file_path != file_path and self.archive.file_path.stem != str(
+            file_path.parent
+        ):
             title = f"{self.archive.file_path}/" + title
         page = f"({self.archive.i+1}/{len(self.archive)}):"
 
@@ -1301,13 +1504,13 @@ class SaltViewer(tk.Tk):
             ".xbm",
         ]:
             return self.open_image(file_path, data)
-        elif suffix in [".tiff"]:
-            # can have multi images
-            pass
+        # elif suffix in [".tiff"]:
+        #    # can have multi images
+        #    pass
         elif suffix in [".svg"]:
             return self.open_svg(file_path, data)
         elif suffix in [".zip", ".rar", ".7z", ".pdf"]:
-            self.open(file_path, data)
+            return self.open(file_path, data)
         else:
             logger.debug(f"Not supported.:{suffix}")
             return None
@@ -1349,7 +1552,8 @@ class SaltViewer(tk.Tk):
 
     def mainloop(self):
         super().mainloop()
-        self.archive.close()
+        if self.archive is not None:
+            self.archive.close()
 
 
 class Icon:
